"""Module for optimizing functions using a :class:`Swarm`."""
# from fragile.optimize.models import ESModel
from fragile.core.swarm import FunctionMapper
from fragile.optimize.env import Function, MinimizerWrapper
